#include <stdio.h>
#include "serialib.h"
#include <iostream>
#include <stdint.h>
#include <time.h>

#define BOOTLOADER_BAUDRATE 1000000
#define BOOTLOADER_ADDRESS  (0x3c00U)

#if defined (_WIN32) || defined( _WIN64)
#define         DEVICE_PORT             "COM3"                               // COM1 for windows
#endif

#ifdef __linux__
#define         DEVICE_PORT             "/dev/ttyS0"                         // ttyS0 for linux
#define Sleep(t)    usleep((t) * 1000UL)
#endif

#define MAX_HEX_SIZE    (16*1024)

using namespace std;

uint8_t hexfile_data[MAX_HEX_SIZE+1];
uint16_t hexfile_max_address;

serialib LS;

// Takes 2 hex numbers and makes a value
unsigned char hex2bin(char *hexstring)
{
    unsigned char ret = 0;
    if((hexstring[0] >= '0') && (hexstring[0] <= '9')) ret = hexstring[0] - '0';
    if((hexstring[0] >= 'a') && (hexstring[0] <= 'f')) ret = hexstring[0] - 'a' + 10;
    if((hexstring[0] >= 'A') && (hexstring[0] <= 'F')) ret = hexstring[0] - 'A' + 10;
    ret <<= 4;
    if((hexstring[1] >= '0') && (hexstring[1] <= '9')) ret += hexstring[1] - '0';
    if((hexstring[1] >= 'a') && (hexstring[1] <= 'f')) ret += hexstring[1] - 'a' + 10;
    if((hexstring[1] >= 'A') && (hexstring[1] <= 'F')) ret += hexstring[1] - 'A' + 10;
    return(ret);
}

int read_hexfile(char *filename)
{
    FILE *hexfile;
    hexfile_max_address = 0;
    for(int i=0;i<MAX_HEX_SIZE;i++) hexfile_data[i] = 0xff;

    hexfile = fopen(filename,"r");
    if(hexfile != NULL) {
        char line[1024];
        while(fgets(line,sizeof(line), hexfile) != NULL)
        {
            int linelen = strlen(line);
            if((linelen > 10) && (line[0] == ':'))
            {
                char * pLine = line+1;
                unsigned int len1, addr1, type1;
                len1 = hex2bin(pLine); pLine += 2;
                addr1 = (((unsigned int)hex2bin(pLine))<<8) + (unsigned int)hex2bin(pLine+2); pLine += 4;
                type1 = hex2bin(pLine); pLine += 2;
                if((((len1 * 2) + 11) <= linelen) && (type1 == 0) && (addr1+(len1*2) < MAX_HEX_SIZE ))
                {
                    for(int i = 0; i<(int)len1; i++)
                    {
                        hexfile_data[i+addr1] = hex2bin(pLine); pLine += 2;
                        if((i+addr1) > hexfile_max_address) hexfile_max_address = (i+addr1);
                    }

                } else {
                  //  cout << "Drop BC: " << len1 << " Addr: " << addr1 << " T: " << type1 << " | " << pLine << "\n";
                }
            }
            else
            {
                cout << "Invalid line in hexfile: " << line << "\n";
            }
        }
        fclose(hexfile);
        return 0;
    } else {
        return 1;
    }
}

// prints every character in buf if printable, count occurence of countchar in buf
int analyse_rxbuf(char *buf, int len, char countchar)
{
    int i, count = 0;
    for(i=0;i<len;i++) {
        if(*buf == countchar) count++;
        if(*buf >= 32) cout << *buf; else cout << ".";
        buf++;
    }
    cout << " | ";
    return count;
}

int main ( int argc, char *argv[] )
{
  int counter1, buntichcount, bc2;
  char checksum, errorcount = 0;
  // char *argsim[] = {"main.cpp","com3","e:\\test.hex"}; argc = 2; argv = argsim;

  if ( argc != 3 ) // argc should be 2 for correct execution
    // We print argv[0] assuming it is the program name
    cout<<"usage: "<< argv[0] <<" <port> <hex-filename>\n";
  else {
                                                                // Object of the serialib class
    int Ret;                                                                // Used for return values
    char Buffer[128];

    cout<<"1: "<< argv[1] << " | " << argv[2] << "\n";

    Ret=read_hexfile(argv[2]);
    if(Ret)
    {
        cout << "Could not read hex-file.\n";
        return Ret;
    }
    //Sleep(2000);

    // Open serial port
    Ret=LS.Open(argv[1],BOOTLOADER_BAUDRATE);                                       // Open serial link at x bauds
    if (Ret!=1) {                                                           // If an error occured...
        printf ("Error while opening port. Permission problem ?\n");        // ... display a message ...
        return Ret;                                                         // ... quit the application
    }
    printf ("Serial port opened successfully !\n");

    // Try to enter bootloader
    counter1 = 2;
    do {
        LS.ReadAWhile(Buffer,128,300);	// Flush buffer

        // Try it first the direct way...
        Ret=LS.WriteChar(0xEE);
        Ret=LS.WriteChar(0xDD);
        Ret=LS.WriteChar(0x11);
        Ret=LS.WriteChar(0xAA);
        if (Ret!=1) {                                // If the writting operation failed ...
            printf ("Error while writing data\n");   // ... display a message ...
            return Ret;                              // ... quit the application.
        }
        Sleep(200);
        Ret=LS.ReadAWhile(Buffer,128,300);
        if(Ret>=1) analyse_rxbuf(Buffer,Ret,'1');
        if((Ret>=1) && (Ret<=64)) {
            printf("Bootloader entered.\n");
            break;
        }
        printf ("Could not enter bootloader, try reset...\n");

        // try flushing buffers
        for(int i=0; i<520; i++) LS.WriteChar('r');
        Sleep(200);
        LS.WriteChar(0xf8);
        LS.WriteChar(0xf8);
        LS.WriteChar(0xf8);
        LS.WriteChar(0xf8);
        Sleep(400);

    } while(--counter1);
    if(!counter1) {
        printf ("Could not enter bootloader!\n");        // ... display a message ...
        return Ret;
    }

    // Count buntichs
    Ret=LS.WriteChar('n');
    if (Ret!=1) {                                // If the writting operation failed ...
        printf ("Error while writing data\n");   // ... display a message ...
        return Ret;                              // ... quit the application.
    }
    Sleep(20);
    Ret=LS.ReadAWhile(Buffer,128,100);
    if(Ret<1) {
        cout << "Could not count devices!\n";
        return Ret;
    }
    buntichcount = analyse_rxbuf(Buffer,Ret,'o');
    cout << "Device count = " << buntichcount << "\n";

    // Flash pages
    for(int i=0;i<(int)(BOOTLOADER_ADDRESS-512);i+=512)
    {
        char progcmd[3];
        progcmd[0] = 'p';
        progcmd[1] = checksum = i & 0xff; // Low byte first
        progcmd[2] = (i >> 8) & 0xff;
        checksum += progcmd[2];
        Ret=LS.Write(progcmd,3);
        if(Ret!=1) { printf ("Error while writing data\n"); return Ret; }
        Ret=LS.Write(&hexfile_data[i],512);
        for(int j=0;j<512;j++) checksum += hexfile_data[i+j];
        if(Ret!=1) { printf ("Error while writing data\n"); return Ret; }
        Ret=LS.WriteChar(checksum);
        if(Ret!=1) { printf ("Error while writing data\n"); return Ret; }
        Ret=LS.ReadAWhile(Buffer,128,100);
        if(Ret<1) {
            cout << "Could not count devices!\n";
            return Ret;
        }
        bc2 = analyse_rxbuf(Buffer,Ret,'o');
        printf("Address 0x%04x written, devices ok = %d \n", i,bc2);
        if(bc2 != buntichcount) errorcount++;

    }

    // Exit bootloader
    Ret=LS.WriteChar('r');
    if(Ret!=1) { printf ("Error while writing data\n"); return Ret; }
    Sleep(20);
    Ret=LS.ReadAWhile(Buffer,128,20);
    if(Ret<1) {
        cout << "Could not count devices!\n";
        return Ret;
    }
    bc2 = analyse_rxbuf(Buffer,Ret,'o');
    cout << "Device count = " << bc2 << "\n";
    if(bc2 != buntichcount) errorcount++;

    // Close the connection with the device

    LS.Close();

    if(!errorcount) {
        cout << "No errors. Bye.\n";
    } else {
        cout << (int)errorcount << " ERRORS OCCURED! TRY AGAIN...\n";
    }

    return errorcount;
  }
}
